##########################################################################
# $Id: default_updusr.pm,v 1.32 2000/10/02 16:36:39 mengel Exp $
#
# DESCRIPTION: 
#       New and Improved! table file parser, etc.
#       This module contain routines that are user callouts to get various
#       pieces of information. Users are required to edit this file
#	and put in in $UPD_USERCODE or their home directory. Please
#	look at the comments for each subroutine to determine the 
#	pieces of information that need to be returned.
#
#	This file just provides examples, do NOT use as is. You MUST 
#	read and understand each function and modify it to conform to 
#	the configuration for your system. 
#
# MODIFICATIONS:
#
##########################################################################
package updusr;

use FileHandle;
use upderr;
use updups;
use updgbl;
use upduti;
use strict;
use integer;
use Exporter();

use vars qw( 
    @ISA 
    @EXPORT  
    $updusr_table 
    %stop_cycles
);

%stop_cycles = {};

@ISA = qw( Exporter );

# export this !
@EXPORT = qw (
 	updusr_get_db
 	updusr_get_proddir
	updusr_predeclare
	updusr_postdeclare
	updusr_get_tablefile
	updusr_get_upsdir
	);

#
# New table data structure looks like:
#
# matchdata: { key => regexp, key => regexp, ... }
#
# updusr_table: { key => [ [matchdata, value], [matchdata, value], ...]
#

# Implement the usual methods by looking up the answers in the
# table data structure.

my $updusr_read_table = 0;

my $updusr_table = {
   'database'		=> [],
   'prod_dir' 		=> [],
   'unwind_prod_dir'	=> [],
   'ups_dir'		=> [],
   'unwind_ups_dir'	=> [],
   'table_dir'		=> [],
   'unwind_table_dir'	=> [],
   'table_file'		=> [],
   'archive_file'	=> [],
   'unwind_archive_file'=> [],
   'predeclare'		=> [],
   'postdeclare'	=> [],
};

sub dump_table {
    my ($key, $list, @tmp, $pair, $match, $val, $mkey, $mpat, $dbg);

    return unless  (upderr_getverbose() > 3);

    $dbg = "\n";
    while( ($key, $list) = each(%{$updusr_table}) ) {
	 $dbg .=  "--------------\n" . $key .  ":\n";
         @tmp = @{$list};
         while ( @tmp ) {
	    $pair = shift(@tmp);
            ($match,$val) = @$pair;
	    while( ($mkey, $mpat) = each(%{$match}) ) {
		$dbg .= " $mkey =~ /$mpat/ && ";
	    }
            $dbg .= $val . "\n";
         }
    }
    upderr_log(4, $dbg, 0);
}

sub updusr_get_db {
    my ($inst_spec, $top_inst_spec,$flags) = @_;

    return get_from_table($inst_spec, $top_inst_spec, $flags, 'database');
}

sub updusr_get_proddir {
    my ($inst_spec, $top_inst_spec,$flags) = @_;

    return get_from_table($inst_spec, $top_inst_spec, $flags, 
				'prod_dir', 'unwind_prod_dir',
				'unwind_archive_file', 'archive_file');
}


sub updusr_get_tablefile {
    my ($inst_spec, $flags) = @_;

    return get_from_table($inst_spec, $inst_spec, $flags, 
				'table_dir', 
				'unwind_table_dir',
				'table_file' );
}

sub updusr_get_upsdir {
    my ($inst_spec,$flags) = @_;
    return get_from_table($inst_spec, $inst_spec, $flags, 
				'ups_dir' ,
				'unwind_ups_dir',
				);
}

sub updusr_predeclare {
    my ($inst_spec,$flags) = @_;

    return get_from_table($inst_spec, $inst_spec, $flags, 'predeclare');
}

sub updusr_postdeclare {
    my ($inst_spec,$flags) = @_;

    return  get_from_table($inst_spec, $inst_spec, $flags, 'postdeclare');
}

sub updusr_unwind_all {
    my ($inst_spec, $flags) = @_;
    return 0;
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
#
# parse the table, setting $updusr_table...
#

$updusr::ups_env_1 = <<'EOF';
 UPS_PROD_NAME="${UPS_PROD_NAME}";
 UPS_PROD_VERSION="${UPS_PROD_VERSION}";
 UPS_PROD_FLAVOR="${UPS_PROD_FLAVOR}";
 UPS_PROD_DIR="${UPS_PROD_DIR}";
 UPS_UPS_DIR="${UPS_UPS_DIR}";
 UPS_THIS_DB="${UPS_THIS_DB}";
 export UPS_PROD_NAME UPS_PROD_VERSION UPS_PROD_FLAVOR UPS_PROD_DIR UPS_UPS_DIR UPS_THIS_DB
EOF

$updusr::ups_env_2 = <<'EOF';
 unset UPS_PROD_NAME
 unset UPS_PROD_VERSION
 unset UPS_PROD_FLAVOR
 unset UPS_PROD_DIR
 unset UPS_UPS_DIR
 unset UPS_THIS_DB
EOF

sub scorematch {
    my $match = @_[0];
    my $score = 0;
    my ($mkey,$mpat);

    while( ($mkey, $mpat) = each(%{$match}) ) {
	if ( $mpat =~ /\.\*/ ) {
	    # wildcards count for 0;
	    next;
        } else {
	    $score += length($mpat);
	}
    }
    return $score;
}

sub cmpmatch {
    return scorematch(${$b}[0]) - scorematch(${$a}[0]);
}

#
# map names as defined in the file to 
# their real names internal to a in inst_spec.
#
my %keymap = (
   'ups_this_db'	=>   'database',
   'ups_prod_dir' 	=>   'prod_dir' ,
   'unwind_prod_dir'	=>   'unwind_prod_dir',
   'ups_ups_dir'	=>   'ups_dir',
   'unwind_ups_dir'	=>   'unwind_ups_dir',
   'ups_table_dir'	=>   'table_dir',
   'unwind_table_dir'	=>   'unwind_table_dir',
   'ups_table_file'	=>   'table_file',
   'ups_archive_file'	=>   'archive_file',
   'unwind_archive_file'=>   'unwind_archive_file',
);

sub updusr_slurp_table {
     my $saw_file = 0;
     my $fh = new FileHandle;
     my $filename;		# name of file
     my $syntaxerrors = 0;	# errors seen
     my $thismatch;		# current match
     my $groupmatch;		# current match -- group
     my $inaction;		# are we in an action= section
     my $curactionval;		# current action list
     my $curactionname;		# current action name
     my $ingroup;		# are we in a group:...end:
     my $incommon;		# are we in a comon:..end:
     my $lastwasmatch;		# previous statement was a match
     my $key;		
     my $val;
     my $val2;
     my $list;
     my $good;			# this line is syntactically good

     $updusr_read_table = 1;
     upderr_log(2,"Entering updusr_slurp_table\n");


     #backward compatibility hack
     if ( -f "$::ENV{UPD_USERCODE_DIR}/config.upd" ) {
         $filename = "$::ENV{UPD_USERCODE_DIR}/config.upd";
     }

     if ( -f "$::ENV{UPD_USERCODE_DIR}/updconfig") {
          $filename = "$::ENV{UPD_USERCODE_DIR}/updconfig";
     }

     if ( !open( $fh, "<$filename" ) ) {
         upderr_log(0, "Unable to open '$filename'.\n");
	 return 0;
     }

     $thismatch = {};
     $groupmatch = {};
     $lastwasmatch = 0;

     while ( <$fh> ) {

         $good = 0;
	 upderr_log(5, "got line: $_");

         if ( /\A(\s*#.*|\s*)\Z/i ) {
	    # oooh! a comment!
	    upderr_log(5, "its a comment...");
	    next;
         }

	 if ( /\A\s*file\s*=\s*updconfig\s*\Z/i ) {
	    $saw_file = 1;
 	    $lastwasmatch = 0;
            $good = 1;
         }

         #
         # the old parser wanted double quotes on Execute commands...
         #
	 if ( /\A\s*Execute\s*\(\s*"(.*)"\s*,\s*((NO_)?UPS_ENV)\s*\)\s*\Z/i ||
	      /\A\s*Execute\s*\((.*),\s*((NO_)?UPS_ENV)\s*\)\s*\Z/i ) {
	    upderr_log(5, "its an execute line...");
	    if ( !$inaction ) { 
	      upderr_log(0, "Error in config file, Execute statment not in action = xxx section at line $.");
            }
	    $curactionval .= $&;
 	    $lastwasmatch = 0;
            $good = 1;
	    next;
         }
	 # any other sort of line ends the action, so
	 if ($inaction) {
	    upderr_log(5, "Ending an action...");
	    unshift( @{$updusr_table->{$curactionname}}, [$thismatch, $curactionval] );
         }
         if ( /\A\s*group:\s*\Z/i ) {
	    upderr_log(5, "it starts a group...");
            $thismatch = {};
	    $groupmatch = {};
	    $ingroup = 1;
	    $inaction = 0;
 	    $lastwasmatch = 0;
            $good = 1;
	    next;
         }
         if ( /\A\s*common:\s*\Z/i ) {
	    upderr_log(5, "it starts a common...");
	    $incommon = 1;
	    $inaction = 0;
 	    $lastwasmatch = 0;
            $good = 1;
	    next;
         }
         if ( /\A\s*end:\s*\Z/i ) {
	    upderr_log(5, "it ends group...");
	    if ( !$ingroup ) {
	      upderr_log(0, "Error in config file, mismatched END: at line $. of updconfig");
	      $syntaxerrors++;
            }
	    $inaction = 0;
	    $ingroup = 0;
 	    $lastwasmatch = 0;
            $good = 1;
	    next;
         }
	 if ( /\A\s*(product|flavor|qualifiers|options|dist_database|dist_node)\s*=\s*(.*?)\s*\Z/i ) {
	    upderr_log(5, "match line for $1\n");
	    $inaction = 0;
	    $key = $1;
	    $val = $2;

	    if (! $lastwasmatch ) {
	        # starts a new chunk
		$thismatch = {};
            }

 	    $lastwasmatch = 1;

	    $key =~ tr[A-Z][a-z];

	    if ( $val eq "ANY" ) {
		# ANY is a wildcard...
	        $val = '.*';
            } else {
		$val =~ s/\W/\\$&/go;
		# colons are multiple item match lists
	        $val =~ s/\\:/|/go;
            }

	    if (!$ingroup && defined($thismatch->{$key})) {
		upderr_log(0,"multiple $key lines not in group:..end: at line $. of updconfig");
		$syntaxerrors++;
	    }
	    if (defined ($groupmatch->{$key})) {
                $thismatch->{$key} = $val;
                $groupmatch->{$key} .= "|" . $val;
 	    } else {
                $thismatch->{$key} = $val;
                $groupmatch->{$key} = $val;
            }
            $good = 1;
         }
	 if ( /\A\s*(ups_this_db|(ups|unwind)_(prod|ups|table|archive)_(dir|file))\s*=\s*"([^"]*)"\s*\Z/i ) {
	    $key = $1;
	    $val = $5;
	    $key =~ tr[A-Z][a-z];
            $key = $keymap{$key};


	    if ($incommon) {
		unshift( @{$updusr_table->{$key}}, [$groupmatch, $val] );
	    } else {
		unshift( @{$updusr_table->{$key}}, [$thismatch, $val] );
            }
 	    $lastwasmatch = 0;
            $good = 1;
         }

	 if ( /\A\s*action\s*=\s*((pre|post)declare)\s*\Z/i ) { 
	    upderr_log(5,"starts an action\n");
	    $inaction = 1;
	    ($curactionname = $1) =~ tr{A-Z}{a-z};
 	    $curactionval = '';
 	    $lastwasmatch = 0;
            $good = 1;
         }
         if (!$good) {
	    upderr_log(0,"Unknown statement type at line $. in updconfig");
	    $syntaxerrors++;
         }
     }

     upderr_log(4,"Before Sorting:");
     dump_table();

     while( ($key, $list) = each(%{$updusr_table}) ) {
	my @tmp  = sort cmpmatch @{$list};
	$updusr_table->{$key} = \@tmp;
     }

     upderr_log(4,"After Sorting:");
     dump_table();
     return $syntaxerrors == 0;
}

sub get_prod_dir_prefix {
    my ($inst_spec) = @_;
    my ($opt, $p, $cmd);

    # dig PROD_DIR_PREFIX out of that database...

    if (defined $inst_spec->{database} && $inst_spec->{database} ne '' ) {
	$opt = "-z $inst_spec->{database}";
    } else {
        $opt = "";
    }

    $cmd = "ups list ${opt} -K prod_dir_prefix 2>/dev/null";
    $p = `$cmd`;
    $p =~ s/\s.*//so;
    upderr_log(3,"list [$cmd] got |$p|\n");

    if ( $p eq '' || $p eq '""' ) {
	# XXX this is untill 'ups list -K prod_dir_prefix' with
	# XXX no other options works.

	$p = `ups list -a ${opt} -K prod_dir_prefix ups 2>/dev/null |  head -1`;
	$p =~ s/\s.*//so;
        upderr_log(3,"second list got $p\n");
	if ( $p eq '""' || $p eq '' ) {
	   $p = `ups list -a ${opt} -K prod_dir_prefix 2>/dev/null | head -1`;
	   $p =~ s/\s.*//so;
           upderr_log(3,"third list got $p\n");
	}
    }
    $p =~ s/"([^"]*)".*/$1/s;

    $::ENV{PROD_DIR_PREFIX} = $p;
    upderr_log(3,"Got PROD_DIR_PREFIX of $::ENV{PROD_DIR_PREFIX}\n");
}

sub get_from_table {
    my ($inst_spec, $top_inst_spec, @keywords) = @_;
    my @pairs;
    my ($pair, $match, $val);
    my $keyword;
	
    # compute UPS_DASH_QUALIFIERS UPS_DASH_FLAVOR, etc.

    ($inst_spec->{qualifiers_dash} = $inst_spec->{qualifiers}) =~ s{\W}{-}go;
    ($inst_spec->{flavor_dash} = $inst_spec->{flavor}) =~ s{\W}{-}go;
    ($inst_spec->{base_flavor} = $inst_spec->{flavor}) =~ s{\+.*}{}o;
    $inst_spec->{options} = $g_flags{opt_O};
    $::ENV{SUFFIX} = (upduti_split_archive_file("ftp://localhost/$inst_spec->{'dist_archive_file'}"))[5];
    if ( !defined($::ENV{SUFFIX} )) {
        $::ENV{SUFFIX} = "tar"
    }
   
    get_prod_dir_prefix( $inst_spec );

    unless ( $updusr_read_table ) {
        updusr_slurp_table();
    }

    upderr_log(4,"get_from_table(@_)\n");
    foreach $keyword (@keywords) {

        next unless (defined($updusr_table->{$keyword}));

	@pairs = @{$updusr_table->{$keyword}};
        OUTER:
	while (@pairs) {
	   $pair = shift(@pairs);
	   ($match,$val) = @$pair;

           INNER:
	   foreach( keys(%{$match}) ) {
	      upderr_log(3,"get_from_table: $keyword -- $match->{$_} -> $val\n");
	      if ( $inst_spec->{$_} =~ /$match->{$_}/ ) {
		 ;
	      } else {
		 next OUTER;
	      }
	   }
	   upderr_log(4,"before expansion: $val\n");

	   $val =~ s/([\\]?\$)(\{[A-Z_]+\})/eval(fixvars($1,$2,$top_inst_spec))/ego;

	   upderr_log(4,"after replacement: $val\n");
	   if ( $keyword =~ /(pre|post)declare/ ) {
	       $val =~ s/Execute\s*\(\s*\"?(.*?)\"?\s*,\s*NO_UPS_ENV\s*\)/$1;/igo;
	       $val =~ s/Execute\s*\(\s*\"?(.*?)\"?\s*,\s*UPS_ENV\s*\)/$updusr::ups_env_1;$1;$updusr::ups_env_2;/igo;
	       upderr_log(5,"about to system($val)");
	       system($val);
           } else {
	       $inst_spec->{$keyword} = $val;
           }
           last OUTER;
	}
    }
    return $inst_spec;
}


#
# turn variabe names in the table file into variable names we use
# internally
#
sub fixvars {
    my ($pre,$string, $top_inst_spec) = @_;
    my ($l, $r);
    my ($dotilde);
    
    my $orig = $string;

    foreach ($string) {

    if ($pre eq "\\\$" || $string eq "ACTION" ) {
	$_ = $pre . $_;
	last;
    }
    if ( /\AUNWIND_/ ){ 
	$dotilde = 1;
    } else {
	$dotilde = 0;
    }

    upderr_log(6,": $_\n");
    s/\{(.*)\}/$1/;
    if ( /\A(_.*)/ )		{ $_ = $1; tr[A-Z][a-z]; $_ = fv2($_); last; }
    if ( /UPS_THIS_DB/ )      	{ $_ = 'database'; $_ = fv2($_); last; }
    if ( /UPS_PROD_NAME/ )      { $_ = 'product'; $_ = fv2($_); last; }
    if ( /UPS_([A-Z]+_DIR)/ )   { $_ = $1; tr[A-Z][a-z]; $_ = fv2($_); last; }
    if ( /UPS_([A-Z]+_FILE)/ )  { $_ = $1; tr[A-Z][a-z]; $_ = fv2($_); last; }
    if ( /UPS_PROD_([A-Z]+)/ )  { $_ = $1; tr[A-Z][a-z]; $_ = fv2($_); last; }
    if ( /DASH_PROD_([A-Z]+)/ ) { $_ = $1."_dash"; tr[A-Z][a-z]; $_ = fv2($_); last; }
    if ( /UPS_DASH_([A-Z]+)/ ) { $_ = $1."_dash"; tr[A-Z][a-z]; $_ = fv2($_); last; }
    if ( /(UNWIND_[A-Z]+_DIR)/ ){ $_ = $1; tr[A-Z][a-z]; $_ = fv2($_); last; }
    if ( /(UNWIND_[A-Z]+_FILE)/ ){ $_ = $1; tr[A-Z][a-z]; $_ = fv2($_); last; }
    if ( /UPS_(BASE_FLAVOR)/ )  { $_ = $1; tr[A-Z][a-z]; $_ = fv2($_); last; }

    s/.*/\$::ENV{$&}/;
    }
    # others are just environment variables
    upderr_log(6, "converted to $string\n");

    if ( $string =~ /\A\$/ ) {
       if (! eval("defined($string)") ) {
	   if ( $string =~ m/\$top_inst_spec->\{(.*)}/ && !$stop_cycles{$1} ) {

	       # this bit of A-B recursion finds the variable if
	       # it is defined in the table file.
	       # regardless what order it was specified.
	       # we set $stop_cycles{x} 1, to prevent infinite
	       # recursion -- the test that got us into this if()
	       # will fail for this value now.

	       $stop_cycles{$1} = 1;
	       get_from_table($top_inst_spec,$top_inst_spec,$1);
	       $stop_cycles{$1} = 0;
           }
       }
       if (! eval("defined($string)") ) {
	   upderr_log(0, "ERROR: In updconfig file, variable \$$orig used before set");
       }
    }
    # do ~user expansion
    if ( $dotilde && $string =~ m/\A~/ ) {
	$string = glob($string);
    }
    return $string;
}

sub fv2 {
    my $string = $_[0];
    my $res;
    $res = '$top_inst_spec->{' . $string . '}';
    return $res;
}

sub moduletest {
    my $inst_spec = { 	'product' => 'perl',
			'version' => 'v5_004',
			'flavor' => 'SunOS+5', 
			'qualifiers' => '', 
			'dist_archive_file' => '/ftp/prd/perl/v5_004/SunOS+5/per.tar',
			'dist_database' => '/ftp/upsdb',
		};
			
    $g_flags{opt_h} = "fnkits.fnal.gov";

    upduti_print_instance_spec(0, $inst_spec);
    upderr_get(); upderr_clear();

    updusr_get_db($inst_spec,$inst_spec, "");
    upderr_get(); upderr_clear();

    updusr_get_proddir($inst_spec,$inst_spec, "");
    upderr_get(); upderr_clear();

    updusr_get_tablefile ($inst_spec, "");
    upderr_get(); upderr_clear();

    updusr_get_upsdir($inst_spec, "");
    upderr_get(); upderr_clear();

    updusr_predeclare($inst_spec, "");
    upderr_get(); upderr_clear();

    updusr_postdeclare($inst_spec, "");
    upderr_get(); upderr_clear();

    upduti_print_instance_spec(0, $inst_spec);
    upderr_get(); upderr_clear();

    $inst_spec = { 	'product' => 'perl',
			'version' => 'v5_004',
			'flavor' => 'IRIX+5', 
			'qualifiers' => '', 
                        'dist_database' => '/tmp/bogus',
			'dist_archive_file' => '/ftp/prd/perl/v5_004/IRIX+5/per.tar',
			'qualifiers' => 'foo:bar:baz',
			};
			
    upduti_print_instance_spec(0, $inst_spec);
    upderr_get(); upderr_clear();

    updusr_get_db($inst_spec,$inst_spec, "");
    upderr_get(); upderr_clear();

    upduti_print_instance_spec(0, $inst_spec);
    upderr_get(); upderr_clear();
}

if ( $0 =~ /(.*)updusr(.*)\.pm/ ) {
    $::ENV{UPD_USERCODE_DIR} = '.'; 
    upderr_setverbose(6);
    moduletest
}

1;
__END__
}
